# frozen_string_literal: true

require 'json'
require 'logger'
require 'socket'
require 'thread'
require 'stringio'

require_relative 'core/controller'
require_relative 'core/config/daemon_config'
require_relative 'core/net/server'
require_relative 'core/util/util'
require_relative 'core/formatter'
require_relative 'core/item'
require_relative 'core/module'

module Zashoku

  class<<self
    attr_accessor :controllers, :server
    def command(message)
      Zashoku.logger.debug("command got #{message}")
      message = message.keys.map(&:to_sym).zip(message.values).to_h

      obj =
        if controllers.key?(message[:mod]) && controllers[message[:mod]].respond_to?(message[:meth])
          controllers[message[:mod]].send(message[:meth], *message[:args])
        else
          nil
        end
      if message[:raw]
        Zashoku.logger.info('responding with unencoded stream')
        if obj.respond_to?(:gets)
          obj
        else
          StringIO.new(obj.to_s)
        end
      else
        Zashoku::Util.encode_object(obj)
      end
    end
  end

  class Daemon
    def initialize(port:)
      @port = port
      @semaphore = Mutex.new

      Zashoku.logger = Logger.new(STDOUT)
      Zashoku.logger.level = Zashoku::CConf[:app][:log_level][:daemon]
      Zashoku.logger.info("server using zashoku v#{Version.join('.')}")
      # Load conf
      Zashoku.conf = DaemonConfig.new
      # Load modules
      Zashoku.modules = Zashoku::Module.load(Zashoku::CConf[:app][:modules][:daemon])
      # Update conf with module defaults
      Zashoku.conf.reload

      Zashoku.controllers =
        Zashoku.modules.keys.zip(Zashoku.modules.values.map(&:controller)).to_h

      Zashoku.controllers.each_value { |c| c.add_observer(self) }
    end

    def listen
      lay_traps!
      @server = Net::Server.new(@port)
      @server.handler = method(:respond)
      Zashoku.logger.info('server listening')
      sleep
    end

    def update(e)
      Zashoku.logger.debug("controller #{e['sender']} sent update #{e['event']}")
      @server.event(e)
    end

    def respond(message)
      Zashoku.logger.debug("daemon.respond #{message}")
      case message['msg']
      when 'modules'
        Zashoku.modules.map { |m, _| m }
      when 'items'
        Zashoku.command(mod: message[:mod], meth: :items)
      when 'fwd'
        Zashoku.command(message)
      when 'helo'
        Zashoku.logger.info("Client saying helo #{message}")
        if message['zv'] == Zashoku::Version && message['app'] == Zashoku::CConf[:app][:name]
          'ok'
        else
          'rejected'
        end
      when 'disconnect'
        nil
      when 'stop'
        cleanup
      when 'up?'
        true
      else
        'what?'
      end
    end

    def lay_traps!
      Signal.trap('INT') do
        Thread.new do
          @semaphore.synchronize { cleanup }
        end
      end
    end

    def cleanup
      @server.exit
      exit
    end
  end
end
